'use strict';

function install$3(VTU, compatConfig) {
  [VTU.config.plugins.DOMWrapper, VTU.config.plugins.VueWrapper].forEach((pluginHost) =>
    pluginHost.install((wrapper) => {
      const { attributes } = wrapper;
      return {
        attributes(attr) {
          const originalAttributes = attributes.call(wrapper, attr);

          if (attr === "value" && compatConfig.WRAPPER_ATTRIBUTES_VALUE) {
            return "value" in wrapper.element ? wrapper.element.value : originalAttributes;
          }

          if (attr === "disabled" && compatConfig.WRAPPER_ATTRIBUTES_DISABLED) {
            return typeof originalAttributes === "string" ? "disabled" : originalAttributes;
          }

          if (attr) {
            return originalAttributes;
          }

          const normalizedDisabled = originalAttributes.disabled === "" ? "disabled" : originalAttributes.disabled;
          const normalizedAttributes = { ...originalAttributes };
          if (compatConfig.WRAPPER_ATTRIBUTES_VALUE && "value" in wrapper.element) {
            normalizedAttributes.value = wrapper.element.value;
          }
          if (compatConfig.WRAPPER_ATTRIBUTES_DISABLED && "disabled" in originalAttributes) {
            normalizedAttributes.disabled = normalizedDisabled;
          }

          return normalizedAttributes;
        },
      };
    })
  );
}

function install$2(VTU, compatConfig) {
  VTU.config.plugins.VueWrapper.install((wrapper) => {
    const { props } = wrapper;
    return {
      props(prop) {
        const originalProps = props.call(wrapper, prop);
        if (!compatConfig.WRAPPER_UNWRAP_PROPS) {
          return originalProps;
        }

        if (prop === undefined) {
          return new Proxy(originalProps, {
            get(target, key) {
              const value = target[key];
              // eslint-disable-next-line no-underscore-dangle
              return value?.__v_raw ?? value;
            },
          });
        }

        // eslint-disable-next-line no-underscore-dangle
        return originalProps?.__v_raw ?? originalProps;
      },
    };
  });
}

class WrapperArray extends Array {
  constructor(results) {
    super();
    Object.assign(this, results);
    this.wrappers = results;
  }

  at(i) {
    return this[i];
  }

  trigger(...params) {
    this.wrappers.forEach((w) => w.trigger(...params));
  }

  exists() {
    return this.length > 0 && this.wrappers.every((x) => x.exists());
  }
}

function install$1(VTU, config) {
  [VTU.config.plugins.DOMWrapper, VTU.config.plugins.VueWrapper].forEach((pluginHost) =>
    pluginHost.install((wrapper) => {
      const { findAll, findAllComponents } = wrapper;
      return {
        findAll: (...args) => {
          const results = findAll.call(wrapper, ...args);

          if (config.WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS) {
            const componentResults = findAllComponents.call(wrapper, ...args);
            results.forEach((v, i) => {
              const matchingComponent = componentResults.find((w) => w.element === v.element);
              if (matchingComponent) {
                results[i] = matchingComponent;
              }
            });
          }

          return config.WRAPPER_FIND_ALL ? new WrapperArray(results) : results;
        },
        findAllComponents: (...args) => {
          const results = findAllComponents.call(wrapper, ...args);
          return config.WRAPPER_FIND_ALL ? new WrapperArray(results) : results;
        },
      };
    })
  );
}

function install(VTU, config) {
  [VTU.config.plugins.DOMWrapper, VTU.config.plugins.VueWrapper].forEach((pluginHost) =>
    pluginHost.install((wrapper) => {
      const { find, findAllComponents, findComponent } = wrapper;
      return {
        find: (...args) => {
          const result = find.call(wrapper, ...args);
          if (!result.exists()) {
            return result;
          }

          if (config.WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS) {
            const componentResults = args[0].ref
              ? [findComponent.call(wrapper, ...args)]
              : findAllComponents.call(wrapper, ...args);

            const matchingComponent = componentResults.find((w) => w.exists() && w.element === result.element);
            if (matchingComponent) {
              return matchingComponent;
            }
          }
          return result;
        },
        findComponent(selector) {
          const result = findComponent.call(wrapper, selector);
          if (
            !result.exists() &&
            selector.ref &&
            config.WRAPPER_FIND_COMPONENT_BY_REF_RETURNS_DOM &&
            wrapper.vm.$refs?.[selector.ref]
          ) {
            return new VTU.DOMWrapper(wrapper.vm.$refs?.[selector.ref]);
          }
          return result;
        },
      };
    })
  );
}

function makeGetOrSetStorageFn(storage) {
  return function getOrSet(id, definition) {
    if (arguments.length === 1) {
      return storage[id];
    }
    // eslint-disable-next-line no-param-reassign
    storage[id] = definition;
    return undefined;
  };
}

function createLocalVue() {
  const plugins = [];
  const directives = {};
  const components = {};
  const mixins = [];

  return {
    extend(options) {
      return options;
    },
    use(...args) {
      plugins.push(args);
    },
    component: makeGetOrSetStorageFn(components),
    directive: makeGetOrSetStorageFn(directives),
    mixin(entry) {
      mixins.push(entry);
    },
    getLocalVueConfig() {
      return {
        ...(plugins.length ? { plugins } : {}),
        ...(Object.keys(directives).length ? { directives } : {}),
        ...(Object.keys(components).length ? { components } : {}),
        ...(mixins.length ? { mixins } : {}),
      };
    },
  };
}

const capitalize = (str) => str.charAt(0).toUpperCase() + str.slice(1);

function smartMerge(...args) {
  const filteredArgs = args.filter((a) => a != null);
  if (filteredArgs.length === 0) {
    return args[0];
  }

  return Object.assign({}, ...filteredArgs);
}

function listenersToProps(listeners) {
  if (listeners == null) {
    return listeners;
  }

  return Object.fromEntries(
    Object.keys(listeners).map((key) => {
      const fixedKey = `on${capitalize(key)}`;
      return [fixedKey, listeners[key]];
    })
  );
}

function normalizeStubs(stubs) {
  if (stubs == null) {
    return stubs;
  }

  const stubsAsObj = Array.isArray(stubs) ? Object.fromEntries(stubs.map((key) => [key, true])) : stubs;
  return Object.fromEntries(
    Object.entries(stubsAsObj).map(([key, stub]) => {
      if (typeof stub === "string") {
        return [key, { name: key, template: stub }];
      }

      return [key, stub];
    })
  );
}

function mergeGlobal(newGlobal, oldGlobal = {}) {
  return {
    components: smartMerge(newGlobal.components, oldGlobal.components),
    config: smartMerge(newGlobal.config, oldGlobal.config),
    directives: smartMerge(newGlobal.directives, oldGlobal.directives),
    mixins:
      /* istanbul ignore next */
      newGlobal.mixins || oldGlobal.mixins ? [...(newGlobal.mixins ?? []), ...(oldGlobal.mixins ?? [])] : undefined,
    mocks: smartMerge(newGlobal.mocks, oldGlobal.mocks),
    plugins:
      /* istanbul ignore next */
      newGlobal.plugins || oldGlobal.plugins ? [...(newGlobal.plugins ?? []), ...(oldGlobal.plugins ?? [])] : undefined,
    provide: smartMerge(newGlobal.provide, oldGlobal.provide),
    renderStubDefaultSlot: oldGlobal.renderStubDefaultSlot,
    stubs: smartMerge(newGlobal.stubs, normalizeStubs(oldGlobal.stubs)),
  };
}

function normalizeConfigOption(obj) {
  const resultObj = Object.fromEntries(Object.entries(obj).filter(([, v]) => v != null));
  return Object.keys(resultObj).length > 0 ? resultObj : null;
}

function normalizeScopedSlots(scopedSlots, scopedSlotsThis) {
  if (scopedSlots == null) {
    return scopedSlots;
  }

  return Object.fromEntries(
    Object.entries(scopedSlots).map(([k, v]) => {
      let normalizedValue = v;
      if (typeof v === "string" && !v.startsWith("<template")) {
        normalizedValue = `<template #default="props">${v}</template>`;
      } else if (typeof v === "function") {
        normalizedValue = (...args) => v.call(scopedSlotsThis, ...args);
      } else {
        throw new Error("Unknown slot type");
      }

      return [k, normalizedValue];
    })
  );
}

function contextOnToProps(on) {
  if (on == null) {
    return on;
  }
  return Object.fromEntries(
    Object.keys(on).map((key) => {
      const capitalized = capitalize(key);
      return [`on${capitalized}`, on[key]];
    })
  );
}

function normalizeProvide(provide) {
  return typeof provide === "function" ? provide() : provide;
}

function normalizeMountArgs(args, config, vueH) {
  if (args == null) {
    return args;
  }

  const {
    // VTU v1 props
    localVue,

    context,
    mocks,
    provide,
    listeners,
    scopedSlots,
    stubs,
    components,
    directives,

    // VTU v2 props
    props,
    propsData,
    slots,
    global,
    ...otherArgsOptions
  } = args;

  const scopedSlotsThis = config.MOUNT_ARGS_SCOPED_SLOTS_THIS
    ? {
        $createElement: vueH,
        $set: (obj, path, value) => {
          // eslint-disable-next-line no-param-reassign
          obj[path] = value;
        },
      }
    : {};

  const {
    plugins,
    directives: localVueDirectives,
    components: localVueComponents,
    mixins,
  } = localVue?.getLocalVueConfig() ?? {};
  const computedArgs = normalizeConfigOption({
    props: smartMerge(
      config.MOUNT_ARGS_CONTEXT_ATTRS ? context?.attrs : null,
      config.MOUNT_ARGS_CONTEXT_PROPS ? context?.props : null,
      config.MOUNT_ARGS_CONTEXT_CLASS && context?.class ? { class: context?.class } : null,
      config.MOUNT_ARGS_CONTEXT_ON ? contextOnToProps(context?.on) : null,
      config.MOUNT_ARGS_LISTENERS ? listenersToProps(listeners) : null,
      propsData,
      props
    ),
    slots: smartMerge(
      config.MOUNT_ARGS_SCOPED_SLOTS ? normalizeScopedSlots(scopedSlots, scopedSlotsThis) : null,
      config.MOUNT_ARGS_CONTEXT_CHILDREN && context?.children ? { default: context.children } : null,
      slots
    ),
    global: normalizeConfigOption(
      mergeGlobal(
        {
          stubs: config.MOUNT_ARGS_STUBS
            ? {
                ...normalizeStubs(stubs),
                ...(config.MOUNT_ARGS_DIRECTIVES && directives
                  ? Object.fromEntries(Object.entries(directives).map(([k, v]) => [`v${capitalize(k)}`, v]))
                  : {}),
              }
            : null,
          mocks: config.MOUNT_ARGS_MOCKS ? mocks : null,
          provide: config.MOUNT_ARGS_PROVIDE ? normalizeProvide(provide) : null,
          plugins,
          mixins,
          components: smartMerge(localVueComponents, config.MOUNT_ARGS_COMPONENTS ? components : null),
          directives: smartMerge(localVueDirectives, config.MOUNT_ARGS_DIRECTIVES ? directives : null),
        },
        global
      )
    ),
  });
  return smartMerge(computedArgs, otherArgsOptions);
}

const installedConfigs = new WeakMap();

function needsNormalization(config) {
  return (
    config.EXPORT_CREATE_LOCAL_VUE || Object.entries(config).some(([k, v]) => k.startsWith("MOUNT_ARGS") && Boolean(v))
  );
}

const compatFlags = Object.freeze({
  EXPORT_CREATE_LOCAL_VUE: "EXPORT_CREATE_LOCAL_VUE",
  EXPORT_CREATE_WRAPPER: "EXPORT_CREATE_WRAPPER",
  EXPORT_ERROR_WRAPPER: "EXPORT_ERROR_WRAPPER",
  EXPORT_WRAPPER_ARRAY: "EXPORT_WRAPPER_ARRAY",

  GLOBAL_STUBS: "GLOBAL_STUBS",

  MOUNT_ARGS_COMPONENTS: "MOUNT_ARGS_COMPONENTS",
  MOUNT_ARGS_CONTEXT_ATTRS: "MOUNT_ARGS_CONTEXT_ATTRS",
  MOUNT_ARGS_CONTEXT_CHILDREN: "MOUNT_ARGS_CONTEXT_CHILDREN",
  MOUNT_ARGS_CONTEXT_CLASS: "MOUNT_ARGS_CONTEXT_CLASS",
  MOUNT_ARGS_CONTEXT_ON: "MOUNT_ARGS_CONTEXT_ON",
  MOUNT_ARGS_CONTEXT_PROPS: "MOUNT_ARGS_CONTEXT_PROPS",
  MOUNT_ARGS_DIRECTIVES: "MOUNT_ARGS_DIRECTIVES",
  MOUNT_ARGS_LISTENERS: "MOUNT_ARGS_LISTENERS",
  MOUNT_ARGS_MOCKS: "MOUNT_ARGS_MOCKS",
  MOUNT_ARGS_PROVIDE: "MOUNT_ARGS_PROVIDE",
  MOUNT_ARGS_SCOPED_SLOTS: "MOUNT_ARGS_SCOPED_SLOTS",
  MOUNT_ARGS_SCOPED_SLOTS_THIS: "MOUNT_ARGS_SCOPED_SLOTS_THIS",
  MOUNT_ARGS_STUBS: "MOUNT_ARGS_STUBS",

  WRAPPER_ATTRIBUTES_DISABLED: "WRAPPER_ATTRIBUTES_DISABLED",
  WRAPPER_ATTRIBUTES_VALUE: "WRAPPER_ATTRIBUTES_VALUE",
  WRAPPER_UNWRAP_PROPS: "WRAPPER_UNWRAP_PROPS",
  WRAPPER_DESTROY: "WRAPPER_DESTROY",
  WRAPPER_DO_NOT_INCLUDE_NATIVE_EVENTS_IN_EMITTED: "WRAPPER_DO_NOT_INCLUDE_NATIVE_EVENTS_IN_EMITTED",
  WRAPPER_DO_NOT_INCLUDE_HOOK_EVENTS_IN_EMITTED: "WRAPPER_DO_NOT_INCLUDE_HOOK_EVENTS_IN_EMITTED",
  WRAPPER_FIND_ALL: "WRAPPER_FIND_ALL",
  WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS: "WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS",
  WRAPPER_FIND_COMPONENT_BY_REF_RETURNS_DOM: "WRAPPER_FIND_COMPONENT_BY_REF_RETURNS_DOM",
  WRAPPER_SET_VALUE_DOES_NOT_TRIGGER_CHANGE: "WRAPPER_SET_VALUE_DOES_NOT_TRIGGER_CHANGE",
  WRAPPER_VUE_SET_VALUE_USES_DOM: "WRAPPER_VUE_SET_VALUE_USES_DOM",
});

const fullCompatConfig = Object.freeze({
  ...Object.fromEntries(Object.keys(compatFlags).map((option) => [option, true])),
});

function installCompat(VTU, compatConfig, vueH = null) {
  if (!VTU) {
    throw new Error("Unknown module for installation");
  }

  if (!compatConfig) {
    throw new Error("compatConfig is required");
  }

  const invalidOptions = Object.keys(compatConfig).filter((option) => !Object.keys(compatFlags).includes(option));
  if (invalidOptions.length > 0) {
    throw new Error(`Got unknown compat options: ${invalidOptions.join(", ")}`);
  }

  if (compatConfig.MOUNT_ARGS_SCOPED_SLOTS_THIS && !compatConfig.MOUNT_ARGS_SCOPED_SLOTS) {
    throw new Error("MOUNT_ARGS_SCOPED_SLOTS_THIS require MOUNT_ARGS_SCOPED_SLOTS compat to function");
  }

  if (compatConfig.MOUNT_ARGS_SCOPED_SLOTS_THIS && !vueH) {
    throw new Error("vueH (third parameter) is required when MOUNT_ARGS_SCOPED_SLOTS_THIS is used");
  }

  const installedCompatConfig = installedConfigs.get(VTU);
  if (installedCompatConfig && installedCompatConfig !== JSON.stringify(compatConfig)) {
    throw new Error(
      "You are trying to install compat layer to vue-test-utils, but it was already installed with different config"
    );
  }
  installedConfigs.set(VTU, JSON.stringify(compatConfig));

  if (compatConfig.GLOBAL_STUBS) {
    // eslint-disable-next-line no-param-reassign
    VTU.config.stubs = VTU.config.global.stubs;
  }

  if (compatConfig.EXPORT_CREATE_WRAPPER) {
    // eslint-disable-next-line no-param-reassign
    VTU.createWrapper = (vm) => (vm instanceof HTMLElement ? new VTU.DOMWrapper(vm) : new VTU.VueWrapper(null, vm));
    // eslint-disable-next-line no-param-reassign
    VTU.Wrapper = VTU.BaseWrapper;
    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.createWrapper = VTU.createWrapper;
      // eslint-disable-next-line no-param-reassign
      VTU.default.Wrapper = VTU.Wrapper;
    }
  }

  if (compatConfig.EXPORT_CREATE_LOCAL_VUE) {
    // eslint-disable-next-line no-param-reassign
    VTU.createLocalVue = createLocalVue;
    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.createLocalVue = VTU.createLocalVue;
    }
  }

  if (compatConfig.EXPORT_ERROR_WRAPPER) {
    // eslint-disable-next-line no-param-reassign
    VTU.ErrorWrapper = function ErrorWrapper(wrapperType) {
      return new Proxy(this, {
        get(obj, prop) {
          switch (prop) {
            case "then":
              // allows for better errors when wrapping `find` in `await`
              // https://github.com/vuejs/test-utils/issues/638
              return undefined;
            case "exists":
              return () => false;
            default:
              throw new Error(`Cannot call ${String(prop)} on an empty ${wrapperType}.`);
          }
        },
      });
    };

    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.ErrorWrapper = VTU.ErrorWrapper;
    }
  }

  if (compatConfig.EXPORT_WRAPPER_ARRAY) {
    // eslint-disable-next-line no-param-reassign
    VTU.WrapperArray = WrapperArray;
    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.WrapperArray = VTU.WrapperArray;
    }
  }

  if (compatConfig.WRAPPER_DESTROY) {
    VTU.config.plugins.VueWrapper.install((wrapper) => ({
      destroy: () => wrapper.unmount(),
    }));
    // eslint-disable-next-line no-param-reassign
    VTU.enableAutoDestroy = VTU.enableAutoUnmount;
    // eslint-disable-next-line no-param-reassign
    VTU.disableAutoDestroy = VTU.disableAutoUnmount;
    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.enableAutoDestroy = VTU.enableAutoDestroy;
      // eslint-disable-next-line no-param-reassign
      VTU.default.disableAutoDestroy = VTU.disableAutoDestroy;
    }
  }

  if (compatConfig.WRAPPER_FIND_ALL || compatConfig.WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS) {
    install$1(VTU, compatConfig);
  }

  if (
    compatConfig.WRAPPER_FIND_BY_CSS_SELECTOR_RETURNS_COMPONENTS ||
    compatConfig.WRAPPER_FIND_COMPONENT_BY_REF_RETURNS_DOM
  ) {
    install(VTU, compatConfig);
  }

  if (compatConfig.WRAPPER_DO_NOT_INCLUDE_NATIVE_EVENTS_IN_EMITTED) {
    // eslint-disable-next-line no-param-reassign
    VTU.VueWrapper.prototype.attachNativeEventListener = () => {};
  }

  if (compatConfig.WRAPPER_DO_NOT_INCLUDE_HOOK_EVENTS_IN_EMITTED) {
    VTU.config.plugins.VueWrapper.install((wrapper) => {
      const { emitted } = wrapper;
      return {
        emitted(event) {
          const result = emitted.call(this, event);
          if (arguments.length === 1) {
            return result;
          }

          Object.keys(result).forEach((k) => {
            if (k.startsWith("hook:")) {
              delete result[k];
            }
          });
          return result;
        },
      };
    });
  }

  if (needsNormalization(compatConfig)) {
    const originalMount = VTU.mount;
    // eslint-disable-next-line no-param-reassign
    VTU.mount = function patchedMount(component, args) {
      return originalMount.call(this, component, normalizeMountArgs(args, compatConfig, vueH));
    };

    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.mount = VTU.mount;
    }

    const originalShallowMount = VTU.shallowMount;
    // eslint-disable-next-line no-param-reassign
    VTU.shallowMount = function patchedMount(component, args) {
      return originalShallowMount.call(this, component, normalizeMountArgs(args, compatConfig, vueH));
    };

    if (VTU.default) {
      // eslint-disable-next-line no-param-reassign
      VTU.default.shallowMount = VTU.shallowMount;
    }
  }

  if (compatConfig.WRAPPER_ATTRIBUTES_VALUE || compatConfig.WRAPPER_ATTRIBUTES_DISABLED) {
    install$3(VTU, compatConfig);
  }

  if (compatConfig.WRAPPER_UNWRAP_PROPS) {
    install$2(VTU, compatConfig);
  }

  if (compatConfig.WRAPPER_VUE_SET_VALUE_USES_DOM) {
    VTU.config.plugins.VueWrapper.install((wrapper) => ({
      setValue(value) {
        return new VTU.DOMWrapper(wrapper.element).setValue(value);
      },
    }));
  }

  if (compatConfig.WRAPPER_SET_VALUE_DOES_NOT_TRIGGER_CHANGE) {
    let blockTriggerringChange = false;
    VTU.config.plugins.DOMWrapper.install((wrapper) => {
      const { trigger, setValue } = wrapper;
      return {
        trigger(event, ...args) {
          if (blockTriggerringChange && event === "change") {
            return null;
          }

          return trigger.call(wrapper, event, ...args);
        },
        setValue(...args) {
          blockTriggerringChange = true;
          try {
            return setValue.call(wrapper, ...args);
          } finally {
            blockTriggerringChange = false;
          }
        },
      };
    });
  }
}

exports.compatFlags = compatFlags;
exports.fullCompatConfig = fullCompatConfig;
exports.installCompat = installCompat;
